open CursorAPI

module Editor = functor(C : CursorAPI.CURSOR) ->
  struct
    (* utility *)
    let movable f c = try ignore (f c); true with _ -> false
    let to_string c = try match C.get c with
      | Data(s)   -> s 
      | Node(n,a) -> "["^n^"]"
           with _ -> "#end#"

    (* show the path to current node *)
    let pwd c =
      let rec pwd_i c =
        if movable C.up c then pwd_i (C.up c);
        print_string ("/"^to_string c)
      in
        pwd_i c; print_endline ""

    (* show the content of current node *)
    let ls c =
      if not (movable C.down c) then
        print_endline "#error#"
      else
        let c = C.down c in
          let rec ls_i c =
            let s = to_string c in if s = "#end#" then () else
              (print_endline s; ls_i (C.next c))
          in
            ls_i c

    (* show current node recursively *)
    let dig c =
      pwd c;
      let rec show d c =
        let s = to_string c in
          if s = "#end#" then () else (
            print_endline (String.make d ' '^s);
            (try show (d+2) (C.down c) with _ -> ());
            show d (C.next c)
          )
      in
        (try show 2 (C.down c) with _ -> ())

    (* generate new node *)
    let gen_node str =
      if Str.string_match (Str.regexp "^\\[.*\\]$") str 0 then
        CursorAPI.Node(String.sub str 1 (String.length str - 2), [])
      else
        CursorAPI.Data(str)

    let rec main c =
       print_string ">>> ";
       let argv = Str.split (Str.regexp " +") (read_line ()) in
         try 
           match String.lowercase (List.hd argv), List.tl argv with
           | "exit", []
           | "quit", []
           | "end" , []
           | "q"   , [] -> c
           | "pwd" , [] -> pwd c; main c
           | "ls"  , [] -> ls  c; main c
           | "prev", [] -> main (C.prev c)
           | "next", [] -> main (C.next c)
           | "up"  , [] -> main (C.up   c)
           | "down", [] -> main (C.down c)
           | "remove", []
           | "del",    [] -> main (C.remove c)
           | "inserta", [s] -> main (C.insertA (gen_node s) c)
           | "insertb", [s] -> main (C.insertB (gen_node s) c)
           | "set",     [s] -> main (C.set (gen_node s) c)
           | "dig",     []  -> dig c; main c
           | _ -> print_endline "#unknown command#"; main c
         with
           | _ -> print_endline "#error#"; main c
  end

module XmlEditor   = Editor(XmlCursor);;
module TableEditor = Editor(TableCursor);;

let _ =
  if Array.length Sys.argv = 1 then
     print_endline ("usage: "^Sys.argv.(0)^" filename")
  else
     let ext_of str =
       try  let idx = String.rindex str '.' + 1 in
         String.lowercase (String.sub str idx (String.length str - idx))
       with Not_found -> ""
     in
     let arg1 = Sys.argv.(1) in
     let arg2 = if Array.length Sys.argv = 2 then arg1 else Sys.argv.(2) in
       match ext_of arg1 with
       | "xml" ->
           print_endline "*** welcome to xml_mode ***";
           let fi = open_in arg1                                   in
           let cr = XmlEditor.main (XmlCursor.input_xml_cursor fi) in
           let  _ = close_in fi                                    in
           let fo = open_out arg2                                  in
              XmlCursor.output_xml_cursor fo cr;
              close_out fo;
       | "csv" ->
           print_endline "*** welcome to csv_mode ***";
           let fi = open_in arg1                                         in
           let cr = TableEditor.main (TableCursor.input_table_cursor fi) in
           let  _ = close_in fi                                          in
           let fo = open_out arg2                                        in
              TableCursor.output_table_cursor fo cr;
              close_out fo;
       | _ ->
           print_endline "*** welcome to debug_mode ***";
