open CursorAPI

(********************************************************************
  module XmlCursor
    an instance of cursor API implemented by the Slit data structure
 ********************************************************************)

(* Slit for XML tree *)

module Xml =
  struct
    type item  = string
    type label = xName * xAttribute list
  end

module XmlSlit = Slit.Make(Xml)

(* fulfill requirements of cursor API *)

type cursor = XmlSlit.slit
let prev    = XmlSlit.prev
let next    = XmlSlit.next
let up      = XmlSlit.up
let down    = XmlSlit.down
let remove  = XmlSlit.remove_elem
let get c   = match XmlSlit.get_elem c with
               | XmlSlit.Leaf(s)       -> CursorAPI.Data(s)
               | XmlSlit.Node((n,a),_) -> CursorAPI.Node(n,a)

let convert_impl d =
  match d with
    | CursorAPI.Data(s)   -> XmlSlit.Leaf(s)
    | CursorAPI.Node(n,a) -> XmlSlit.Node((n,a),[])

let insertA d = XmlSlit.insert_after  (convert_impl d)
let insertB d = XmlSlit.insert_before (convert_impl d)
let set d c = match d with
               | CursorAPI.Data(s)   -> XmlSlit.set_elem (XmlSlit.Leaf(s)) c
               | CursorAPI.Node(n,a) -> XmlSlit.set_elem (XmlSlit.Node((n,a),
                   match XmlSlit.get_elem c with
                     | XmlSlit.Leaf(s)    -> []
                     | XmlSlit.Node(_,ch) -> ch
                   )) c

(********************************************************************
  xml cursor IO
     input_xml_cursor :  in_channel -> cursor
    output_xml_cursor : out_channel -> cursor -> unit
 ********************************************************************)

module CursorIoImpl =
  struct
    open Yaxpo
    open Yaxpodom

    let string_of_channel fin =
      let len = in_channel_length fin in
      let buf = String.create len in
        really_input fin buf 0 len; buf

    let yaxpo_of_string text =
      let reader   = Cps_reader.make_utf8_string_reader text in
      let (_,root) = build_doc_tree reader in
         !root

    let tree_of_yaxpo y =
      let rec process_ele (Element(qn, atts, cnts)) =
       let conv_name (QName(_,localname,_)) =
         localname
       in
          let conv_atts atts =
         List.map (fun (Att(qn,vref)) -> (conv_name qn, !vref)) !atts
       in
       let conv_cnts cnts =
         List.map iter
           (List.filter (function Text(_)|CDATA(_)|Subele(_)->true | _->false) !cnts)
       in
         XmlSlit.Node((conv_name qn, conv_atts atts), conv_cnts cnts)
     and iter = function
       | Text(s)
       | CDATA(s)  -> XmlSlit.Leaf(s)
       | Subele(e) -> process_ele e
     in
       process_ele y

    let output_tree ch =
      let rec output_elem = function
        | XmlSlit.Leaf(s)        -> output_string ch s
        | XmlSlit.Node((n,a),[]) ->
            output_char   ch '<';
            output_string ch n;
            output_attrs     a;
            output_string ch "/>";
        | XmlSlit.Node((n,a),children) ->
            output_char   ch '<';
            output_string ch n;
            output_attrs     a;
            output_char   ch '>';
            output_elems     children;
            output_string ch "</";
            output_string ch n;
            output_char   ch '>';
      and output_attr (nm,vl) =
            output_char   ch ' ';
            output_string ch nm;
            output_string ch "=\"";
            output_string ch vl;
            output_char   ch '"';
      and output_attrs al = List.iter output_attr al
      and output_elems el = List.iter output_elem el
      in  output_elem
  end


let input_xml_cursor ch =
  XmlSlit.slit_of_tree
    (CursorIoImpl.tree_of_yaxpo
      (CursorIoImpl.yaxpo_of_string
        (CursorIoImpl.string_of_channel ch)))

let output_xml_cursor ch cur =
  CursorIoImpl.output_tree ch
    (XmlSlit.tree_of_slit cur)
